/**
   @file    QualityExpressionsProfilerPAPI.h
   @ingroup QualityExpressionProfilerPAPI
   @brief   Quality Expression profiling headers for PAPI Counters
   @author  Laurent Morin
   @verbatim
   Revision:       $Revision$
   Revision date:  $Date$
   Committed by:   $Author$

   This file is part of the Periscope performance measurement tool.
   See http://www.lrr.in.tum.de/periscope for details.

   Copyright (c) 2005-2014, Technische Universitaet Muenchen, Germany
   See the COPYING file in the base directory of the package for details.

   @endverbatim
 */

#ifndef QUALEXPRSEMANTIC_PAPI_DEF

#ifndef QUALITYEXPRESSION_PROFILERPAPI_H_
#define QUALITYEXPRESSION_PROFILERPAPI_H_

#include "quality-expressions/QualityExpressionsProfiler.h"

/* ---------------------------------------------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------------------------------------------- */

/**
 * @defgroup QualityExpressionProfilerPAPI Profiling system from PAPI counters.
 * @ingroup QualityExpressionProfiler
 */

/** @def QE_PROFILER_PAPI_BASE
    @brief Define the starting range for the semantic.
    @ingroup QualityExpressionProfilerPAPI
 */
#define QE_PROFILER_PAPI_BASE           0x20000

/** @def QE_PROFILER_PAPI_MASK
    @brief Define the mask valid for the port range local semantic.
    Note that we have no clear view of how many counters could be available in the future, so the range is quite wide.
    We will also introduce subnets for a selective management of these counters.
    @ingroup QualityExpressionProfilerPAPI
 */
#define QE_PROFILER_PAPI_MASK           0xFFFF

/** @brief Define the PAPI kind of counter of info.
    @ingroup QualityExpressionProfilerPAPI
 */
typedef enum qualexpr_kind_papi_t {
    PAPI_KIND_UNDEF,
    PAPI_KIND_INFO,
    PAPI_KIND_PRESET,
} qualexpr_kind_papi_t;

/** @brief Define the PAPI event list.
    @ingroup QualityExpressionProfilerPAPI
 */
typedef enum qualexpr_event_papi_t {
    QE_PROFILER_PAPI_UNDEFINED = QE_PROFILER_PAPI_BASE,         //!< Default event undefined.

#define QUALEXPRSEMANTIC_PAPI_DEF( DefName, KIND )      \
    QE_PROFILER_PAPI_ ## DefName,
#include "quality-expressions/QualityExpressionsProfilerPAPI.h"
#undef QUALEXPRSEMANTIC_PAPI_DEF
    QE_PROFILER_PAPI_NOMORE                                     //!< List end.
} qualexpr_event_papi_t;

#ifdef __cplusplus
extern "C" {
#endif

semantic_namespace_t qualExpr_papi_registerListener( listen_event_func_t listener );

void qualExpr_papi_unregisterListener( listen_event_func_t listener );

void qualExpr_papi_startCounters( void );

void qualExpr_papi_stopCounters( void );

enum qualexpr_kind_papi_t qualExpr_papi_eventKind( enum qualexpr_event_papi_t semantic );

int qualExpr_papi_isAvailable( enum qualexpr_event_papi_t semantic );

void qualExpr_papi_addSemantic( enum qualexpr_event_papi_t semantic );

void qualExpr_papi_reset( void );

#ifdef __cplusplus
}
#endif

#endif // /QUALITYEXPRESSION_PROFILERPAPI_H_

#else  // --------------------- List section -------------------------------
#ifndef QUALEXPRSEMANTIC_PAPI_DEF_ONLY_PRESET
QUALEXPRSEMANTIC_PAPI_DEF( INF_CORES, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_MAX_MHZ, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_MIN_MHZ, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPUID_FAMILY, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPUID_MODEL, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPUID_STEPPING, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_MODEL, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_NCPU, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_NNODES, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_REVISION, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_SOCKETS, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_CPU_THREADS, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_TOTALCPUS, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( INF_VENDOR, INFO )
QUALEXPRSEMANTIC_PAPI_DEF( PRESET_BASE, UNDEF )
#endif
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_CN, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_MSP, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_NTK, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_PRC, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_TKN, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BR_UCN, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_BTAC_M, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CA_CLN, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CA_INV, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CA_ITV, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CA_SHR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CA_SNP, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CSR_FAL, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CSR_SUC, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_CSR_TOT, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_DP_OPS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FAD_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FDV_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FMA_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FML_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FP_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FP_OPS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FP_STAL, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FPU_IDL, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FSQ_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FUL_CCY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FUL_ICY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_FXU_IDL, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_HW_INT, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_INT_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_DCA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_DCH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_DCM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_DCR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_DCW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_ICA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_ICH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_ICM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_ICR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_ICW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_LDM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_STM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_TCA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_TCH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_TCM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_TCR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L1_TCW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_DCA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_DCH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_DCM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_DCR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_DCW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_ICA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_ICH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_ICM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_ICR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_LDM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_STM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_TCA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_TCH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_TCM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_TCR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L2_TCW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_DCA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_DCH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_DCM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_DCR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_DCW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_ICA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_ICH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_ICM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_ICR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_LDM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_STM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_TCA, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_TCH, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_TCM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_TCR, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_L3_TCW, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_LD_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_LST_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_MEM_RCY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_MEM_SCY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_MEM_WCY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_PRF_DM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_REF_CYC, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_RES_STL, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_SP_OPS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_SR_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_STL_CCY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_STL_ICY, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_SYC_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TLB_DM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TLB_IM, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TLB_SD, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TLB_TL, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TOT_CYC, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TOT_IIS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_TOT_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_VEC_DP, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_VEC_INS, PRESET )
QUALEXPRSEMANTIC_PAPI_DEF( PAPI_VEC_SP, PRESET )
#endif
